VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "OctreeNode"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
'The original C++ code is (c) 2004, Sjaak Priester, Amsterdam.
'VB translation is (c) 2004, A. Skrobov

Option Explicit
Public Red As Long
Public Green As Long
Public Blue As Long
Private Children(0 To 7) As OctreeNode
Public nPrev As OctreeNode, nNext As OctreeNode
Public Pixels As Long
Public PalEntry As Byte
Public Level As Long
Public IsLeaf As Boolean
Public Property Get Child(ByVal Index As Integer) As OctreeNode
    Set Child = Children(Index)
End Property
Public Property Set Child(ByVal Index As Integer, ByVal NewChild As OctreeNode)
    Set Children(Index) = NewChild
End Property
Public Function Clone() As Object
    Set Clone = Me
End Function

Public Sub Create(ByVal Level As Long, ByVal Head As OctreeNode, NumLeaves As Long)
Me.Level = Level
Me.IsLeaf = (Level = 0)
If Me.IsLeaf Then
    NumLeaves = NumLeaves + 1
ElseIf Not (Head Is Nothing) Then
    Me.InsertAfter Head
    Me.UpdatePosition
End If
End Sub

Public Function ReduceBefore(ByVal Node As OctreeNode) As Boolean
    ' true if other should be reduced before we are
    If Me.Level = Node.Level Then
        ReduceBefore = (Me.Pixels < Node.Pixels)
    Else
        ReduceBefore = (Me.Level < Node.Level)
    End If
End Function

Public Sub UpdatePosition()
    ' Determine our priority, in other words our position in the list.
    Dim pNext As OctreeNode: Set pNext = Me.nNext

    ' Increment as long as pNext should be reduced before we are.
    Do While Not (pNext Is Nothing)
        If ReduceBefore(pNext) Then Exit Do
        Set pNext = pNext.nNext
    Loop

    If Not (pNext Is Me.nNext) Then
        Me.Remove               ' Remove us from old position in list...
' THIS LINE IS DAMNED!
'        Me.InsertBefore (pNext) ' ...and insert us at new position.
' /THIS LINE IS DAMNED
        Set Me.nPrev = pNext.nPrev
        Set Me.nNext = pNext
    
        Set Me.nPrev.nNext = Me
        Set Me.nNext.nPrev = Me
    End If
End Sub

Public Sub Remove()
    Set Me.nPrev.nNext = Me.nNext
    Set Me.nNext.nPrev = Me.nPrev
End Sub

Public Sub InsertBefore(ByVal Node As OctreeNode)
    Set Me.nPrev = Node.nPrev
    Set Me.nNext = Node

    Set Me.nPrev.nNext = Me
    Set Me.nNext.nPrev = Me
End Sub

Public Sub InsertAfter(ByVal Node As OctreeNode)
    InsertBefore Node.nNext
End Sub

Public Function ChildIndex(ByVal r As Byte, ByVal g As Byte, ByVal b As Byte) As Integer
Dim rBit As Long, gBit As Long, bBit As Long
    ' The octree trick: calculate child index based on the m_Level'th bits
    ' of the color bytes.
    rBit = (r \ 2 ^ Me.Level) And 1
    gBit = (g \ 2 ^ Me.Level) And 1
    bBit = (b \ 2 ^ Me.Level) And 1

    ChildIndex = (rBit * 4) + (gBit * 2) + bBit
End Function

Public Sub SetColor(ByVal r As Byte, ByVal g As Byte, ByVal b As Byte, ByVal Head As OctreeNode, NumLeaves As Long)
    Me.Pixels = Me.Pixels + 1
    If Me.IsLeaf Then
        ' If we are a leaf node, accumulate color bytes in totals.
        Me.Red = Me.Red + r
        Me.Green = Me.Green + g
        Me.Blue = Me.Blue + b
    Else
        If Not (Me.nNext Is Nothing) Then Me.UpdatePosition
                                        ' Perhaps we should move further down the list,
                                        ' because we represent more pixels.

        ' We are not a leaf node, so delegate to one of our children.
        Dim Index As Integer: Index = Me.ChildIndex(r, g, b)

        ' If we don't have a matching child, create it.
        If Children(Index) Is Nothing Then
            Set Children(Index) = New OctreeNode
            Children(Index).Create Me.Level - 1, Head, NumLeaves
        End If

        Children(Index).SetColor r, g, b, Head, NumLeaves  ' recursion
    End If
End Sub

Public Sub GetQuantized(NumLeaves As Long)
    ' GetQuantized a non-leaf node to a leaf
Dim r As Long, g As Long, b As Long, i As Integer
Dim Child As OctreeNode
    For i = 0 To 7
        Set Child = Children(i)
        If Not (Child Is Nothing) Then

            ' Accumulate color totals of our children...
            r = r + Child.Red
            g = g + Child.Green
            b = b + Child.Blue
    
            ' ...and delete them.
            Set Children(i) = Nothing
    
            NumLeaves = NumLeaves - 1
        End If
    Next
    Me.Red = r
    Me.Green = g
    Me.Blue = b

    ' Now we have become a leaf node ourselves.
    Me.IsLeaf = True
    NumLeaves = NumLeaves + 1
End Sub

Public Sub FillPalette(Palette() As OLE_COLOR, Index As Integer)
Dim i As Integer, r As Byte, g As Byte, b As Byte
    If Me.IsLeaf Then
        ' If we are a leaf, calculate the color by dividing the color totals
        ' through the number of pixels we represent...
        r = Me.Red / Me.Pixels
        g = Me.Green / Me.Pixels
        b = Me.Blue / Me.Pixels

        ' ...and put it in the palette.
        Palette(Index) = r + 256& * g + 65536 * b
        Me.PalEntry = Index
        Index = Index + 1
    Else
        ' Otherwise, ask our child nodes.
        For i = 0 To 7
            If Not (Children(i) Is Nothing) Then _
                Children(i).FillPalette Palette, Index ' recursion
        Next
    End If
End Sub

Public Function GetPaletteIndex(ByVal r As Byte, ByVal g As Byte, ByVal b As Byte) As Byte
    ' If we are a leaf node, the palette index is simply stored.
    If Me.IsLeaf Then
        GetPaletteIndex = Me.PalEntry
    Else
        ' Otherwise, the matching child node must know more about it.
        Dim Index As Integer: Index = Me.ChildIndex(r, g, b)
        GetPaletteIndex = Children(Index).GetPaletteIndex(r, g, b)   ' recursion
    End If
End Function

